// 3 note improv - Jim Bumgardner 2010
//
// A simple algorithmic composition that plays three notes and
// changes them one at a time, choosing new notes that score highly
// given a set of arbitrary (western) criteria.

import themidibus.*; //Import the library
MidiBus myBus; // The MidiBus

boolean isMute = false;

int   kLowestNoteNumber = 32;    // lowest MIDI pitch
int   kHighestNoteNumber = 32+64;

int kMidiChannel = 0;
int kDefaultVelocity = 64-20;  // -20 for piano
int kNbrNotes = 3;
int kDefWidth = 800;
int kDefHeight = 600;

int[] notes;

long  vLastChange;
int kChangeFrequencyLow = 1000/1;   // 1 for strings, 30 for piano...
int  kChangeFrequencyHigh = 2000/1;
int  vChangeDur;

public class Mutation
{
  int  idx;
  int  pitch;
  int score;
  
  Mutation()
  {
      idx = int(random(kNbrNotes));
      pitch = notes[idx]+int(random(11))-5;
      
      int[] n = new int[kNbrNotes];
      for (int i = 0; i < kNbrNotes; ++i)
        n[i] = notes[i];

      n[idx] = pitch;
      
      // score it here...
      score = 0;
      if (pitch < kLowestNoteNumber || pitch > kHighestNoteNumber) // stay in range
        score -= 50;
      if (lastMut != null && idx == lastMut.idx)  // move a different note
        score -= 25;
      if (pitch == notes[idx])  // prefer motion
        score -= 10;
      else if (abs(pitch - notes[idx]) <= 2)
        score += 5;  // prefer stepwise motion

      // prefer motion in highest notes...
      if (idx == 0)
        score -= 2;
      if (idx == 1)
        score -= 1;     

      int minPitch = 200;
      int maxPitch= 0;

      for (int i = 0; i < kNbrNotes; ++i) {  // score newly made intervals
          if (n[i] > maxPitch)
            maxPitch = n[i];
          if (n[i] < minPitch)
            minPitch = n[i];
          if (i == idx)
            continue;

          switch (abs(pitch - n[i]) % 12)
          {
          case 0:  score -= 2;  break;
          case 1:  score -= 10;  break;   // minor 2nd are dissonant
          case 2:  break;
          case 3:  score += 10;  break;
          case 4:  score += 10;  break;
          case 5:  score += 5;  break;
          case 6: score -= 5;  break;      // tritones are yucky
          case 7:  score += 5;  break;
          case 8:  score += 10;  break;
          case 9:  score += 10;  break;
          case 10:  break;
          case 11:  score -= 10;  break;  // minor 2nd
          }
      }
      // score spread
      if (maxPitch - minPitch > 16)
        score -= 30;
      if (maxPitch - minPitch > 12)
        score -= 5;

      // keep fingers sorted
      if (n[0] > n[1])
        score -= 30;
      if (n[0] > n[2])
        score -= 30;
      if (n[1] > n[2])
        score -= 30;

      // prefer larger intervals in bass
      if (n[1] - n[0] < n[2] - n[1])  // prefer larger interval on bottom
        score -= 10;        

    vChangeDur = int(random(kChangeFrequencyLow, kChangeFrequencyHigh));
  }
  
};

Mutation lastMut = null;

void setup()
{
  size(kDefWidth, kDefHeight);
  colorMode(HSB, 1);
  
  notes = new int[kNbrNotes];

  for (int i = 0; i < kNbrNotes; ++i) {
      notes[i] = 64;
  }

  MidiBus.list(); // List all available Midi devices on STDOUT. This will show each device's index and name.
  myBus = new MidiBus(this, "Bus 1", "Java Sound Synthesizer"); // Create a new MidiBus with no input device and the default Java Sound Synthesizer as the output device.
  int stringEnsemble = 49;
  myBus.sendMessage(0xc0,kMidiChannel,stringEnsemble,0);

  smooth();
  
  // rStart = 30 + random(60);

  background(0);
  noStroke();
  frameRate(24);
  
  vLastChange = millis();
  lastMut = new Mutation();
}

void drawNotes()
{
  float noteRange = kHighestNoteNumber - kLowestNoteNumber;
  float pixRange = height - 64;

  for (int i = 0; i < kNbrNotes; ++i)
  {
    fill((notes[i] % 12) / 12.0,1,1);
    int py = int((height-32) - (notes[i]-kLowestNoteNumber)*pixRange/noteRange);
    int px = width*i/kNbrNotes + width/(kNbrNotes*2);
    ellipse(px, py, width/kNbrNotes-64, 32);
  }
}


void genMutations()
{
  Mutation mut = new Mutation();
  for (int i = 0; i < 40; ++i) {
      Mutation mut2 = new Mutation();
      if (mut2.score > mut.score) {
        mut = mut2;
      }
  }
  
  if (!isMute)
      myBus.sendNoteOff(kMidiChannel, notes[mut.idx], 0);
  notes[mut.idx] = mut.pitch;
  if (!isMute)
      myBus.sendNoteOn(kMidiChannel, notes[mut.idx], kDefaultVelocity + int(random(21))-10);

  lastMut = mut;
}

void draw()
{
  // background(0); // try blur here at smaller sizes...
  filter(BLUR);
  // filter(ERODE);
  
  long tm = millis();
  
  if (tm - vLastChange > vChangeDur)
  {
      // change a note
      genMutations();
      vLastChange = tm;
      vChangeDur = int(random(kChangeFrequencyLow, kChangeFrequencyHigh));
  }

  drawNotes();
} 
 
void allNotesOff()
{
  for (int i = 0; i < 128; ++i)
    myBus.sendNoteOff(kMidiChannel, i, 0);
}

void keyPressed() {
  if (key == ' ') 
  {
    isMute = !isMute;
    if (isMute) {
      allNotesOff();
    }
    println("MUTE " + (isMute? "ON" : "OFF"));
  }
}

void stop()
{
  allNotesOff();
  super.stop();
}


